import cnd_io
import yamale
import yaml


class Structure:
    def __init__(self, project, definition, _print, provider=None, branch="main", base_path=''):
        self._print = _print
        self._project = project
        self._definition = definition
        self._branch = branch
        self._base_path = base_path
        if provider is None:
            self._provider = cnd_io.CndProviderLocalfile(creds={}, print=_print)
        else:
            self._provider = provider
        self._cnd_io = cnd_io.CndIO(self._provider, print=_print)
        self._structure = None
        self._schema = None
        self._files = None

    @property
    def structure(self):
        if self._structure is not None:
            return self._structure
        self._structure = self._cnd_io.pull_file(self._project, f'{self._base_path}{self._definition}.yml', branch=self._branch)
        if self._structure is False:
            self._print.info_e(f"Cannot found file for structure : {self._project} | {self._base_path}{self._definition}.yml")
            raise NameError('Structure unreadable')
        return self._structure

    @property
    def schema(self):
        if self._schema is not None:
            return self._schema
        self._schema = self._cnd_io.pull_file(self._project, f'{self._base_path}schema.yml', branch=self._branch)
        if self._schema is False:
            self._print.info_e(f"Cannot found file for schema : {self._project} | {self._base_path}schema.yml")
            raise NameError('Schema unreadable')
        return self._schema

    @property
    def files(self):
        if self._files is not None:
            return self._files
        result = []
        safe_structure = yaml.safe_load(self.structure)
        for file in safe_structure['files']:
            folder = "/".join([self._base_path, safe_structure["source_folder"]]).replace('//','/')
            if folder[0] == '/':
                folder = folder[1:]
            file_content = self._cnd_io.pull_file(self._project, f'{folder}/{file["name"]}', branch=self._branch)
            data = {
                "file_name": file["target"] if "target" in file else file["name"],
                "file_content": file_content,
                "step": file["step"]
            }
            if "target_base_path" in safe_structure:
                data["file_name"] = "/".join([safe_structure["target_base_path"], data["file_name"]])
            if 'repeat' in file:
                data['repeat'] = file['repeat']
            result.append(data)
        return result

    def _validate(self):
        schema = yamale.make_schema(content=self.schema)
        data = yamale.make_data(content=self.structure)
        try:
            yamale.validate(schema, data)
            return True
        except ValueError as e:
            raise NameError('Validation failed')

    def get(self):
        self._validate()
        return self.files


from mockito import when, unstub  # noqa: E402
from expects import *  # noqa: F403, E402
from mamba import description, context, it  # noqa: E402
import yaml  # noqa: E402
import src.cnd_scaffold as scaffold
import tests.vars as vars


with description("File") as self:
    with before.each:
        unstub()
        self.tested_instance = scaffold.file.File(vars._print)

    with context("__init__"):
        with it("should init the print in the class"):
            expect(self.tested_instance._print).to(equal(vars._print))

    with context("apply single file"):
        with before.each:
            self.expected = vars.read_yaml_file('tests/data/single_file/expected.yml')
            source = vars.read_yaml_file('tests/data/single_file/source.yml')
            self.file_name, self.file_content = self.tested_instance.apply(source["file_name"], source["file_content"], source["data"])

        with it("should init the print in the class"):
            expect(self.file_name).to(equal(self.expected["file_name"]))

        with it("should init the print in the class"):
            expect(self.file_content).to(equal(self.expected["file_content"]))

    with context("apply single file with loop"):
        with before.each:
            self.expected = vars.read_yaml_file('tests/data/single_file_loop/expected.yml')
            source = vars.read_yaml_file('tests/data/single_file_loop/source.yml')
            self.file_name, self.file_content = self.tested_instance.apply(source["file_name"], source["file_content"], source["data"])

        with it("should init the print in the class"):
            expect(self.file_name).to(equal(self.expected["file_name"]))

        with it("should init the print in the class"):
            expect(self.file_content).to(equal(self.expected["file_content"]))

    with context("apply single file with loop"):
        with before.each:
            self.expected = vars.read_yaml_file('tests/data/single_file_loop_advanced/expected.yml')
            source = vars.read_yaml_file('tests/data/single_file_loop_advanced/source.yml')
            self.file_name, self.file_content = self.tested_instance.apply(source["file_name"], source["file_content"], source["data"])

        with it("should init the print in the class"):
            expect(self.file_name).to(equal(self.expected["file_name"]))

        with it("should init the print in the class"):
            expect(self.file_content).to(equal(self.expected["file_content"]))


from .__version__ import (__version__)
from .file import File
from .structure import Structure
import cnd_io
import copy
import yaml


class Scaffold:
    step = ['init', 'build', 'runtime']

    def __init__(self, _print):
        self._print = _print
        self._file = File( self._print)

    def _file_apply(self, file_name, file_content, data):
        return self._file.apply(file_name, file_content, data)

    def update_step(self, step):
        self.step = step

    def _clean_file_name(self, file_name, base_path):
        list = [base_path, file_name]
        clean_list = [i for i in list if i is not None]
        return "/".join(clean_list)

    def apply(self, step, model_files, data, base_path=None):
        result = {}
        for file in model_files:
            if step not in file['step']:
                continue
            if 'repeat' not in file:
                _file_name, _file_content = self._file_apply(file['file_name'], file['file_content'], data)
                _file_name = self._clean_file_name(_file_name, base_path)
                result[_file_name] =_file_content
                continue
            for item in data[file["repeat"]]:
                sub_data = copy.deepcopy(data)
                sub_data.update(item)
                _file_name, _file_content = self._file_apply(file['file_name'], file['file_content'], sub_data)
                _file_name = self._clean_file_name(_file_name, base_path)
                result[_file_name] =_file_content
        return result

    def load_model(self, project, definition, provider=None, branch="main", base_path=''):
        structure = Structure(project, definition, self._print, provider=provider, branch=branch, base_path=base_path)
        return structure.get()

    def _get_cnd_io(self, provider):
        if provider is None:
            _provider = cnd_io.CndProviderLocalfile(creds={}, print=self._print)
        else:
            _provider = provider
        return cnd_io.CndIO(_provider, print=self._print)

    def _commit_files(self, files, project, branch):
        for file_name, file_content in files.items():
            self._cnd_io.commit_file(project, file_name, file_content, branch)
        return True

    def push_files(self, files, project, provider=None, branch="main", commit_message="Pushing from CndScaffold"):
        self._cnd_io = self._get_cnd_io(provider)
        self._commit_files(files, project, branch)
        result = self._cnd_io.push_files(project, commit_message=commit_message, branch=branch)
        if result is False:
            self._print.info_e('Cannot push content')
            self._print.trace_e(f'Content: {yaml.dump(self._cnd_io._files)}')
            raise NameError('Content cannot be pushed')
        return True


import jinja2
from .__version__ import (__version__)  # noqa: F401


class File:
    def __init__(self, _print):
        self._print = _print
        self._print.log_d(f"File Class ({__version__})")

    def _jinja_replace(self, content, data):
        environment = jinja2.Environment()
        template = environment.from_string(content)
        return template.render(data)

    def apply(self, file_name, file_content, data):
        _file_name = self._jinja_replace(file_name, data)
        _file_content = self._jinja_replace(file_content, data)
        return _file_name, _file_content


from yamale.yamale_error import YamaleError  # noqa: E402
from mockito import when, unstub  # noqa: E402
from expects import *  # noqa: F403, E402
from mamba import description, context, it  # noqa: E402
import yaml  # noqa: E402
import src.cnd_scaffold.scaffold as scaffold
import tests.vars as vars
import cnd_io


with description("CndScaffold") as self:
    with before.each:
        unstub()
        self.tested_instance = scaffold.Scaffold(vars._print)

    with context("__init__"):
        with it("should init the print in the class"):
            expect(self.tested_instance._print).to(equal(vars._print))

    with context("update_step"):
        with it("should update the step"):
            new_step = ['A', 'B']
            self.tested_instance.update_step(new_step)
            expect(self.tested_instance.step).to(equal(new_step))

    with context("_commit_files"):
        with before.each:
            files = vars.read_yaml_file('tests/data/apply_repeat/expected.yml')
            self.tested_instance._cnd_io = vars._cnd_io
            self.result = self.tested_instance._commit_files(files, "project", "main")

        with it("should update commit list"):
            expect(self.result).to(equal(True))

        with it("should update commit list"):
            expect(len(self.tested_instance._cnd_io._files["project"]["main"])).to(equal(2))

    with context("_commit_files"):
        with it("should return an instance of cnd_io"):
            _provider = cnd_io.CndProviderLocalfile(creds={}, print=vars._print)
            _cnd_io = self.tested_instance._get_cnd_io(_provider)
            expect(isinstance(_cnd_io, cnd_io.CndIO)).to(equal(True))

        with it("should return an instance of cnd_io if provider is not provide"):
            _cnd_io = self.tested_instance._get_cnd_io(None)
            expect(isinstance(_cnd_io, cnd_io.CndIO)).to(equal(True))

    with context("apply single file"):
        with it("should generate file"):
            expected = vars.read_yaml_file('tests/data/apply/expected.yml')
            source = vars.read_yaml_file('tests/data/apply/source.yml')
            files = self.tested_instance.apply('init', source["model_files"], source["data"])
            expect(files).to(equal(expected))

        with it("should generate multiple file"):
            expected = vars.read_yaml_file('tests/data/apply_multiple/expected.yml')
            source = vars.read_yaml_file('tests/data/apply_multiple/source.yml')
            files = self.tested_instance.apply('init', source["model_files"], source["data"])
            expect(files).to(equal(expected))

        with it("should  repeat file"):
            expected = vars.read_yaml_file('tests/data/apply_repeat/expected.yml')
            source = vars.read_yaml_file('tests/data/apply_repeat/source.yml')
            files = self.tested_instance.apply('init', source["model_files"], source["data"])
            expect(files).to(equal(expected))